#include "axml.h"
#include "common.h"
#include "http_comm.h"


int myNum, workers;
double globTime = 0.0;
int permutationSendCount = 0;
int permutationRecvCount = 0;
int numOfPerm = 1;
char *masterAddr;
char *proxyAddr;
int masterPort;
int proxyPort;
int myId;
char myIdStr[10];
int worker_count;
double totalSendTime = 0.0;
int totalSendTries = 0;
clock_t startCpuTime, endCpuTime;
double cpuTime;





int getTreeRanges(int **treeRanges, char **msg)
{
   char *s1, *s2, *s3;
   char *s;
   char *temp;
   int ptr, lower, upper, i, total;
   s = (char *) malloc(32);

   ptr = -1;
   s3 = NULL;

   if (s1 = strstr(*msg, "}#{"))
      if (s2 = strstr(s1 + 2, "}#{"))
         s3 = strstr(s2 + 2, "}#{");

   if (s3)
   {
      s1 = strchr(*msg, '{');
      if (s1)
      {
         s2 = strstr(s1, "}#{");
         s3 = s2;
         ptr = 0;
         strncpy(s, s1, 10);
         sscanf(s, "{%d", &total);
         *treeRanges = (int *) malloc(total * 2 * sizeof(int));
         s1 = strchr(s1, '(');
         while ((s2 = strchr(s1, ')')) && (s2 < s3))
         {
            strncpy(s, s1, s2 - s1 + 1);
            s[s2 - s1 + 1] = 0;
            sscanf(s, "(%x %x)", &lower, &upper);
            (*treeRanges)[ptr * 2] = lower;
            (*treeRanges)[ptr * 2 + 1] = upper;
            ptr++;
            s1 = s2 + 1;
         }
         s1 = strdup(s3 + 2);
         free(*msg);
         *msg = s1;
      }
   } else
      *treeRanges = NULL;
   free(s);
   return ptr;
}


int registerWorker()
{
   struct_http_req *http_req;
   char *buf;
   int status;

   myId = 0;
   http_req = createNewHttpReq(HTTP_GET, masterAddr, masterPort, "/REGISTER", 0);
   buf = sendHttpReq(http_req, &status);
   if ((buf != NULL) && (!status))
   {
      sscanf(buf, "%x", &myId);
      printf("ID assigned: %x\n", myId);
   }
   free(buf);

   if (!myId)
      printf("Unable to reach master. No ID assigned...\n");
   sprintf(myIdStr, "%08x", myId);
   return myId;
}


char *requestSeqFile()
{
   struct_http_req *http_req;
   int status;
   char *buf;

   http_req = createNewHttpReq(HTTP_GET, masterAddr, masterPort, myIdStr, 0);
   buf = sendHttpReq(http_req, &status);
   if (status)
      return NULL;
   else
      return buf;
}

int sendHttpMsg(char **buf)
{
   struct_http_req *http_req;
   char *tmp;
   int status;
   double startTime;

   http_req = createNewHttpReq(HTTP_POST, masterAddr, masterPort, myIdStr, 0);
   if (*buf)
      appendHttpReq(http_req, *buf, strlen(*buf));
   startTime = gettime();
   tmp = sendHttpReq(http_req, &status);
   totalSendTime += gettime() - startTime;
   totalSendTries++;
   if (!status)
   {
      if (*buf)
         free(*buf);
      *buf = tmp;
   }
   return status;
}


double getAvgSendTime()
{
   double result;
   result = totalSendTime / totalSendTries;
   totalSendTime = 0.0;
   totalSendTries = 0;
   return result;
}


int get_args(int argc, char *argv[], boolean print_usage, analdef * adef)
{                               /* get_args */
   int c;
   char *s;
   boolean bad_opt = FALSE;
   int i, time;
   char buf[2048];
   char *s1;


   run_id[0] = 0;
   workdir[0] = 0;
   tree_file[0] = 0;
   masterAddr = NULL;
   masterPort = 0;
   proxyAddr = NULL;
   proxyPort = 0;


   while (!bad_opt && ((c = getopt(argc, argv, "n:w:m:p:")) != -1))
   {
      fprintf(stderr, "%s\n", optarg);
      switch (c)
      {
      case 'n':                /* Name, or ID of run */
         strcpy(run_id, optarg);
         break;
      case 'w':                /* Working directory */
         strcpy(workdir, optarg);
         break;
      case 'm':
         s1 = strchr(optarg, ':');
         s1[0] = '\0';
         s1++;
         masterAddr = strdup(optarg);
         sscanf(s1, "%i", &masterPort);
         break;
      case 'p':
         s1 = strchr(optarg, ':');
         s1[0] = '\0';
         s1++;
         proxyAddr = strdup(optarg);
         sscanf(s1, "%i", &proxyPort);
         break;
      case '?':
      default:
         bad_opt = TRUE;
         break;
      }
   }


/* Set the full path to the working directory. It's up to using functions
 * of workdir to resolve this path, which may contain instances of ./ ../
 * and //, and handle resulting errors. (N.B.: We take care not to introduce
 * double slashes because some implementations of realpath() (notably in AIX)
 * choke on them). */
   if (workdir[0] == 0 || workdir[0] != '/')
   {
      getcwd(buf, sizeof(buf));
      if (buf[strlen(buf) - 1] != '/')
         strcat(buf, "/");
      strcat(buf, workdir);
      if (buf[strlen(buf) - 1] != '/')
         strcat(buf, "/");
      strcpy(workdir, buf);
   }

   if (!masterAddr || !masterPort)
   {
      masterAddr = strdup("localhost");
      masterPort = 5555;
   }

/* Check for errors */
   if (bad_opt)
   {
      if (print_usage)
      {
         s = strrchr(argv[0], '/');
         s = (s == NULL) ? argv[0] : s + 1;
         fprintf(stderr, "Usage: %s [-m <master host>:<port>] [-p <proxy host>:<port>] [-n <run_name>] [-w <work_dir>]\n", s);
         fflush(stderr);
      }
      exit(255);
   }
}                               /* get_args */





/*==========================================================================*/
/*                             "main" routine                               */
/*==========================================================================*/

int main(int argc, char *argv[])
{                               /* DNA Maximum Likelihood */
   analdef *adef;
   rawdata *rdta;
   cruncheddata *cdta;
   tree *tr;                    /*  current tree */

   int j, *perm, l;
   int myprogtype;
   int *prog;
   double workerRLH;
   FILE *permutationFile, *resultFile;
   double startTime;
   char permFileName[1024], resultFileName[1024];
   int parsimonyTrees = 0;
   int ptreesCount = 0;
   int endIt = 0;
   int *work;
   int workLength;
   int workParams = 10;
   int rearrangementsMax, rearrangementsMin;
   char tmp[32];
   char *treeVector;
   int numOfOptimizeJobs, optimizeJobsSent, optimizeJobsReceived, numOfEvaluateJobs, evaluateJobsSent, evaluateJobsReceived;
   int from;
   double lh, recvdlh, optlh, t;
   int i, k, impr, stepwidth;
   jobQueue *jobList;
   int betterFound;
   char *parsimonyTree;

   char *msg;
   struct_http_req *http_req;
   char *treeBuf;
   char *tempStr;
   int jobId, lastJobId;
   struct_jobIdList jobIdList;
   char *jobIdListStr;
   int *treeRanges;
   int treeCount;
   char textOutFileName[1024];
   int optimizeJobCount;
   int retryCount;


   printf("%s %s with %s %s (%s)\n\n", programName, programVersion, http_comm_name, http_comm_version, http_comm_date);
   startTime = gettime();
   cpuTime = 0.0;
   startCpuTime = clock();

   CATEGORIES = 0;
   Multiple = 1;


   if (!(adef = (analdef *) malloc(sizeof(analdef))))
   {
      printf("ERROR: Unable to get memory for analysis definition\n\n");
      return 1;
   }

   if (!get_args(argc, argv, (boolean) 1, adef));

   if (!(rdta = (rawdata *) malloc(sizeof(rawdata))))
   {
      printf("ERROR: Unable to get memory for raw DNA\n\n");
      return 1;
   }

   if (!(cdta = (cruncheddata *) malloc(sizeof(cruncheddata))))
   {
      printf("ERROR: Unable to get memory for crunched DNA\n\n");
      return 1;
   }

   if (proxyAddr && proxyPort)
   {
      printf("Master: %s:%d (via %s:%d)\n", masterAddr, masterPort, proxyAddr, proxyPort);
      setupHttpProxy(proxyAddr, proxyPort);
   } else
      printf("Master: %s:%d (direct)\n", masterAddr, masterPort);

   do
   {
      myNum = registerWorker();
      if (!myNum)
         sleep(5);
   }
   while (!myNum);
   do
   {
      tempStr = requestSeqFile();
      if (!tempStr)
         sleep(5);
   }
   while (!tempStr);
   strcpy(seq_file, tempStr);

   tr = (tree *) malloc(sizeof(tree));

   while (1)
   {
      printf("\nEntering Main loop: %s\n", seq_file);
      INFILE = fopen(seq_file, "r");
      if (!getinput(adef, rdta, cdta, tr))
         return 1;
      fclose(INFILE);

      if (!makeweights(adef, rdta, cdta))
         return 1;
      if (!makevalues(rdta, cdta))
         return 1;
      if (adef->empf && !empiricalfreqs(rdta, cdta))
         return 1;
      reportfreqs(adef, rdta);
      if (!linkdata2tree(rdta, cdta, tr))
         return 1;
      if (!linkxarray(3, 3, cdta->endsite, &freextip, &usedxtip))
         return 1;
      if (!setupnodex(tr))
         return 1;
      if (!setupEqualityVector(tr))
         return 1;
      treeStringLength = (tr->mxtips * (nmlngth + 32)) + 256 + (tr->mxtips * 2);
      workLength = workParams + treeStringLength;
      work = (int *) malloc(workLength * sizeof(int));
      myTreestr = (char *) malloc(treeStringLength);
      bestTreestr = (char *) malloc(treeStringLength);
      perm = (int *) malloc((tr->mxtips + 1) * sizeof(int));
      tl = (treelist *) malloc(sizeof(treelist));

      initList(20);
      jobId = 0;
      jobIdList.max = 100;
      jobIdList.count = 0;
      jobIdList.jobIds = (int *) malloc(jobIdList.max * sizeof(int));
      optimizeJobCount = 0;

      while (1)
      {
         work[8] = SIMPLE_REQUEST;
         msg = setTreeStr(work, 0, NULL);
         if (!sendHttpMsg(&msg))
         {
            lastJobId = jobId;

            treeCount = getTreeRanges(&treeRanges, &msg);
            getTreeStr(msg, work, &jobId, myTreestr);
            free(msg);

            switch (work[8])
            {
            case RANDOM_TREE:
               fprintf(stderr, "RANDOM_TREE %f\n", gettime() - startTime);
               makePermutation(perm, tr);
               RAxML_parsimony(perm, treeStringLength, &parsimonyTree);
               str_treeReadLenSpecial(parsimonyTree, tr);
               free(parsimonyTree);
               treeEvaluate(tr);
               treeString(myTreestr, tr, tr->start->back, 1);
               work[8] = PARSIMONY_TREE;
               msg = setTreeStr(work, jobId, myTreestr);
               retryCount = 0;
               while (sendHttpMsg(&msg) && (retryCount++ < 5))
                  sleep(3);
               if (msg)
                  free(msg);
               break;
            case TREE_OPTIMIZE:
               fprintf(stderr, "TREE_OPTIMIZE %f\n", gettime() - startTime);
               if ((jobId & 0xF0000000) != (lastJobId & 0xF0000000))
               {
                  resetList();
                  jobIdList.count = 0;
                  optimizeJobCount = 0;
               }
               optimizeJobCount++;
               strcpy(bestTreestr, myTreestr);
               treeOptimize(tr, work[0], work[1], work[2], work[3], work[4]);
               if (jobIdList.count == jobIdList.max)
               {
                  jobIdList.max += 100;
                  jobIdList.jobIds = realloc(jobIdList.jobIds, jobIdList.max);
               }
               jobIdList.jobIds[jobIdList.count++] = jobId;
               work[8] = OPTIMIZE_REQUEST;
               msg = setTreeStr(work, jobId, bestTreestr);
               retryCount = 0;
               while (sendHttpMsg(&msg) && (retryCount++ < 5))
                  sleep(3);
               if (msg)
                  free(msg);
               break;
            case COMBINED_OPTIMIZE_EVALUATE:
               fprintf(stderr, "COMBINED_OPTIMIZE_EVALUATE (%d) %f\n", treeCount, gettime() - startTime);
               fflush(stderr);
               strcpy(bestTreestr, myTreestr);
               for (i = 0; i < treeCount; i++)
                  treeOptimize(tr, work[0], work[1], work[2], treeRanges[i * 2], treeRanges[i * 2 + 1]);
               free(treeRanges);
               for (i = 0; i < tl->size; i++)
               {
                  str_treeReadLen(tl->trees[i]->tree, tr);
                  smoothTree(tr, 1);
                  if (CATEGORIES)
                     evaluateCAT(tr, tr->start);
                  else
                     evaluate(tr, tr->start);
                  if (tr->likelihood > lh)
                  {
                     lh = tr->likelihood;
                     treeString(myTreestr, tr, tr->start->back, 1);
                  }
               }

               str_treeReadLen(myTreestr, tr);
               smoothTree(tr, smoothings / 4);
               if (CATEGORIES)
                  evaluateCAT(tr, tr->start);
               else
                  evaluate(tr, tr->start);
               treeString(myTreestr, tr, tr->start->back, 1);
               work[8] = COMBINED_RESULT;
               msg = setTreeStr(work, jobId, myTreestr);
               retryCount = 0;
               while (sendHttpMsg(&msg) && (retryCount++ < 5))
                  sleep(3);
               if (msg)
                  free(msg);
               break;
            case TREE_EVALUATE:
               if (jobIdList.count > 0)
               {
                  fprintf(stderr, "TREE_EVALUATE %f\n", gettime() - startTime);
                  fflush(stderr);
                  lh = unlikely;
                  for (i = 0; i < tl->size; i++)
                  {
                     str_treeReadLen(tl->trees[i]->tree, tr);
                     smoothTree(tr, 1);
                     if (CATEGORIES)
                        evaluateCAT(tr, tr->start);
                     else
                        evaluate(tr, tr->start);

                     if (tr->likelihood > lh)
                     {
                        lh = tr->likelihood;
                        treeString(myTreestr, tr, tr->start->back, 1);
                     }
                  }
                  str_treeReadLen(myTreestr, tr);
                  smoothTree(tr, smoothings / 4);
                  if (CATEGORIES)
                     evaluateCAT(tr, tr->start);
                  else
                     evaluate(tr, tr->start);
                  treeString(myTreestr, tr, tr->start->back, 1);
                  work[8] = TREE_RESULT;
                  treeBuf = setTreeStr(work, jobId, myTreestr);
                  jobIdListStr = (char *) malloc(jobIdList.count * 9 + strlen(treeBuf) + 256);
                  strcpy(jobIdListStr, "{\0");
                  tempStr = (char *) malloc(10);
                  for (i = 0; i < jobIdList.count; i++)
                  {
                     sprintf(tempStr, "%08x ", jobIdList.jobIds[i]);
                     strcat(jobIdListStr, tempStr);
                  }
                  free(tempStr);
                  jobIdListStr[strlen(jobIdListStr) - 1] = '}';
                  strcat(jobIdListStr, "#");
                  strcat(jobIdListStr, treeBuf);
                  retryCount = 0;
                  while (sendHttpMsg(&jobIdListStr) && (retryCount++ < 5))
                     sleep(3);
                  if (jobIdListStr)
                     free(jobIdListStr);
                  free(treeBuf);
                  resetList();
                  jobIdList.count = 0;
                  fprintf(stderr, "   Avg sendTime: %fs    optimizeJobs: %d\n", getAvgSendTime(), optimizeJobCount);
                  endCpuTime = clock();
                  cpuTime += ((double) (endCpuTime - startCpuTime)) / CLOCKS_PER_SEC;
                  startCpuTime = clock();
               } else
                  sleep(5);
               break;
            case FINISH_AXML:
               printf("FINISH_AXML\n");
               goto cleanup;
            default:
               fprintf(stderr, "FATAL ERROR\n");
               exit(1);
            }
         } else
            sleep(3);
      }


    cleanup:
      RAxML_parsimony_cleanup();

      free(perm);
      free(work);
      free(myTreestr);
      free(rdta->y0);
      if (rdta->wgt != NULL)
         free(rdta->wgt);
      if (rdta->wgt2 != NULL)
         free(rdta->wgt2);
      if (rdta->sitecat != NULL)
         free(rdta->sitecat);

      if (cdta->alias != NULL)
         free(cdta->alias);
      if (cdta->aliaswgt != NULL)
         free(cdta->aliaswgt);
      if (cdta->patcat != NULL)
         free(cdta->patcat);
      if (cdta->patrat != NULL)
         free(cdta->patrat);
      if (cdta->wr != NULL)
         free(cdta->wr);
      if (cdta->wr2 != NULL)
         free(cdta->wr2);
      freextip = (xarray *) NULL;
      usedxtip = (xarray *) NULL;
      freeTree(tr);
      goto finish;
   }

 finish:

   fprintf(stderr, "Elapsed time [real Time / CPU Time] %f / %f\n", gettime() - startTime, cpuTime);
   return 0;
}                               /* DNA Maximum Likelihood */
